import os.path

import pytest

import horos_io._utils as _utils


@pytest.mark.parametrize("f", ["cmr.py", "cli.py", "_utils.py"])
def test_glob_ssf(f):
    paths = _utils.globSSF("*", root_dir="horos_io")
    assert f in paths


@pytest.mark.parametrize("i, i_expected", [(9, "09"), pytest.param(-1, ...),
                                           pytest.param(100, ...)])
def test__to_str(i, i_expected):
    if i_expected is not ...:
        assert _utils._to_str(i) == i_expected
    else:
        with pytest.raises(ValueError):
            _utils._to_str(i)


def test__has_dicom(horos_test_seq_paths):
    for horos_test_seq_path in horos_test_seq_paths:
        assert _utils._has_dicom(horos_test_seq_path)
    assert not _utils._has_dicom(".")



def test_get_ids(horos_test_root):
    for ID, target_ID in zip(_utils.get_ids(horos_test_root), ["Impression_Cmr0064", "Impression_Cmr0067"]):
        assert ID == target_ID



import os
from typing import List

import numpy as np
import pandas as pd
import pydicom

from horos_io import load_lax_sequence
from horos_io._utils import _to_str
from horos_io.core import get_n_slices_from_seq_path, get_n_frames_from_seq_path, _get_name_from_template

from horos_io.cmr import Path


def _load_basaL_first_as_list(basal_first_file: Path) -> List[str]:
    basal_df = pd.read_csv(basal_first_file, dtype=str)
    return list(basal_df["basal_first"])


def load_sequence(path_to_sequence: Path, basal_first: bool) -> np.ndarray:
    """
    Legacy variant, relying on user input for ordering

    convenience function to let the algorithm decide, if it is lax or sax;
    will need to pass basal_first though, just in case...
    even though

    a problem is, is that sequences can be repeated multiple times,
    the first code in IM-XXXX-0001.dcm or IM-XXXX-0001-0001.dcm denotes this;

    in our data we have only 1 sequence of however many repetitions there were
    Args:
        path_to_sequence:
        basal_first: pass a dummy boolean if you load a LAX sequence; otherwise for SAX a boolean should be passed,
        that denotes, whether in the SAX sequence lower slice number means more basal; will then reorder, such that
        Apex is first
    Returns:
    """
    p = path_to_sequence
    return load_lax_sequence(p) if get_n_slices_from_seq_path(p) == 1 else load_sax_sequence(p, basal_first)


def load_sax_sequence(path_to_sequence: Path, basal_first: bool) -> np.ndarray:
    """
        returns a numpy array, where the contents of the array are pydicom.FileDataset of the images of shape
    (n_frames, n_slices)
    Args:
        path_to_sequence:
        basal_first: if True, will invert the slice order
    Returns:
    """
    n_frames = get_n_frames_from_seq_path(path_to_sequence)
    n_slices = get_n_slices_from_seq_path(path_to_sequence)
    ordering = -1 if basal_first else 1

    return np.array([[
        pydicom.dcmread(os.path.join(path_to_sequence,
                                     _get_name_from_template(path_to_sequence,
                                                             fr"IM-\d\d\d\d-00{_to_str(f + 1)}-00{_to_str(s + 1)}.dcm")))
        for s in range(n_slices)[::ordering]]
        for f in range(n_frames)])



"""
Console script for horos io;
will incorporate functionality for running visual sanity checks on contours overlayed on top of the data;

generating .csv files for data
"""
import os
import sys
import typing
from typing import Optional, NoReturn

import click
import pandas as pd
from matplotlib import pyplot as plt
from tqdm import tqdm

from horos_io import load_horos_contour
from horos_io._utils import globSSF, mask_from_omega_contour, get_log, get_log_path
from horos_io.cmr import Path, get_image_info, get_contour_info, get_combined_info
from horos_io.ui import shitty_manual_confirm
from horos_io.validation import visually_confirm_omega_iter, write_log, last_validation_was_successful, contour_is_valid


@click.group()
def main():
    """Console script for horos_io."""
    return 0


@click.command("image_info",
               help="saves information about the horos dataset as csv, such that it can be later used for efficient"
                    "loading of the data")
@click.option("--root", default=".", help="directory to the Horos Dataset")
@click.option("--silent", default=False, is_flag=True, help="if set, will not print out the result")
@click.option("--out", default=None,
              help="how the image_info,csv should be saved; if not specified, will save an image_info.csv in root")
def make_image_info_csv(root: Path,
                        out: Optional[Path],
                        silent: bool) -> NoReturn:
    """
    one-time operation to update the information of the dataset;
    columns:
    ("seq_path",  <relative path starting from the root of the data>
     "location",  joins the rootdir with seq path
     "ID",        <Impression_Cmr{ID}>
     "slice_type", one of config.slice_types
     "n_frames",
     "n_slices",

    seq_path is relative to the root
    location is joint path of root and seq_path

    creates a csv file containing location of the sequence.
    sequence can be loaded with load_sequence
    Args:
        silent: if silent, will not print out the result to console
        root: root of the data; contents should be directories of Impression Studies
        out: optional parameter to specify, what and where to save; if None, will save image_info.csv inside root
    Returns:
    """
    click.echo(f"retrieving image info from {root}")
    result = get_image_info(root)
    if not silent:
        click.echo("Findings:\n\n")
        with pd.option_context('display.max_rows', None, 'display.max_columns',
                               None):  # more options can be specified also
            click.echo(result)
    if out is None:
        out = os.path.normpath(os.path.join(root, "image_info.csv"))
    click.echo(f"saving result to {out}")
    result.to_csv(out)


@click.command("contour_info",
               help="saves information about the horos dataset as csv, such that it can be later used for efficient"
                    "loading of the data")
@click.option("--root", default=".", help="directory to the Horos Dataset")
@click.option("--silent", default=False, is_flag=True, help="if set, will not print out the result")
@click.option("--out", default=None,
              help="how the image_info,csv should be saved; if not specified, will save an image_info.csv in root")
def make_contour_info_csv(root: Path, out: Optional[Path], silent: bool) -> NoReturn:
    """
    one-time operation to update the information of the dataset;

    ("ID": <Impression_Cmr{ID}>,
     "contour_path": relative path to the contour file, starting from root of the data,
     "contour_type": one of config.contour_types,
     "location": joins the root dir with the sequence path
     "slice_type": one of config.slice_types,
    )

    creates a csv file containing location of all contour files; or whether they exist at all.
    Args:
        root: root of the data; contents should be directories of Impression Studies
    Returns:
    """
    root = os.path.normpath(root)
    click.echo(f"creating contour info for {root}")
    result = get_contour_info(root)
    if not silent:
        click.echo("Findings:\n\n")
        with pd.option_context('display.max_rows', None, 'display.max_columns',
                               None):  # more options can be specified also
            click.echo(result)

    if out is None:
        out = os.path.normpath(os.path.join(root, "contour_info.csv"))
    click.echo(f"saving contour info to: {out}")
    result.to_csv(out)


@click.command()
@click.option("--by", required=True, help="the name that will be recorded in the log_")
@click.option("--root", default=".", help="source root of the Horos dataset")
@click.option("--only_unconfirmed", is_flag=True, default=False,
              help="will only prompt you cases, that have not been confirmed by the user yet (registered in log_)")
@click.option("--log", default=None,
              help="if None, will save or update a val_contour_log.csv in root, else checks your specified log_ instead")
def validate(by: str, root: Path, only_unconfirmed: bool, log: Path):
    """
    visually confirm literally every existing omega contour... (might need to change with params to return
    an adequate iterator)

    will also save a dataframe / update a dataframe in <root>/val_contour_log.csv,
    that logs review result with timestamp
    """
    click.echo(f"starting validation process in root: {root}")

    for ID, omega, f, s, cines, contours in tqdm(visually_confirm_omega_iter(combined_info=get_combined_info(root))):
        if only_unconfirmed and last_validation_was_successful(get_log(log, root), ID=ID, frame=f, slice=s,
                                                               contour_type=omega):  # CAVE: hard coded
            continue
        fig, ax = plt.subplots()
        ax.imshow(cines[f, s].pixel_array)
        ax.imshow(mask_from_omega_contour(cines, contours, f, s), cmap="jet", alpha=0.5)
        ax.figure.suptitle(f"{ID}: {omega} f: {f}, s: {s}, ")
        ok, remark = shitty_manual_confirm(ax)
        write_log(log, **locals())
    click.echo("batch finished validation!")


@click.command("rename", help="renames all contours with .xml and .rois_series attached to it")
@click.argument("old", nargs=1)
@click.argument("target", nargs=1)
@click.option("--root", help="Data root")
def rename_contour_files(root: str, old: str, target: str) -> typing.NoReturn:
    """
    renames .xml and .roi_series files by dictionary, that lie in the impression folders:
    root |--- Impression_CmrXXXX
                    |---- lv_epi.xml
                    |---- lv_epi.roi_series
    Args:
        root:
    Returns:
    """
    files = globSSF(f"*/{old}.xml", root_dir=root) + globSSF(f"*/{old}.rois_series", root_dir=root)
    for f in files:
        path, f_name = os.path.split(f)
        f_name, f_ending = f_name.split(".")
        src = os.path.normpath(os.path.join(root, path, f"{old}.{f_ending}"))
        dst = os.path.normpath(os.path.join(root, path, f"{target}.{f_ending}"))
        if os.path.isfile(dst):
            raise FileExistsError(f"{dst} for {src} already exists")
        os.rename(src, dst)
        click.echo(f"renamed {src} to {dst}")
    click.echo("finished")


@click.group("check", help="given a root and a log file, checks if the user (--by) has seen all the samples")
def check():
    pass


@click.option("--by", required=True, help="the name to check for in the logs")
@click.option("--root", default=".", help="source root to the Horos dataset")
@click.option("--log", default=None,
              help="if None, will check for val_contour_log.csv in root, else checks your specified log instead")
@click.command("failed", help="check to log for which contours did not and in validation by the user")
def check_failed(by: str, root: str, log: str):
    click.echo(
        f"checking for failed validations by {by} for dataset in {root}, using logfile {get_log_path(log, root)}")
    log_df = get_log(log, root)
    failed = log_df.apply(lambda row: not last_validation_was_successful(log_df,
                                                                         by=row["by"],
                                                                         ID=row["ID"],
                                                                         frame=row["frame"],
                                                                         slice=row["slice"]), axis=1)
    result = log_df[(log_df["by"] == by) & failed]
    result = result[["by", "ID", "frame", "slice"]].drop_duplicates()
    with pd.option_context('display.max_rows', None, 'display.max_columns', None):  # more options can be specified also
        click.echo(result)
        click.echo("----------------")
        click.echo(f"These samples did not end with confirmation by user {by}")


@click.command("unobserved",
               help="looks at the data and the logs to determine, which contours by which ID still needs to,"
                    "be evalutated by a given user;")
@click.option("--root", default=".", help="source root to the Horos dataset")
@click.option("--log", default=None,
              help="if None, will check for val_contour_log.csv in root, else checks your specified log instead")
@click.option("--by", required=True, help="the name to check for in the logs")
def check_unobserved(by: str, root: str, log: str):
    click.echo(
        f"checking, which contours have not been seen by user {by} yet in datasat at {root}, using log {get_log_path(log, root)}")
    log_df = get_log(log, root)
    combined_info_ = get_combined_info(root)
    # combined_info_["ID", "contour_type"]
    click.echo("Under construction")
    return

    # for g_n, group in combined_info_.groupby(["ID", "slice_type", "contour_type", "location_contour", "location_images"]):
    #     click.echo(group)
    #
    # # TODO: open the contour to look for which frame / slice needs to be seen
    # # TODO: filter by omega


@click.command("existing", help="look at exsiting contours in root, given a number of contour types; will check"
                                "for existing <contour_type>.xml files that are in child directories of root")
@click.argument("contour_types", nargs=-1)
@click.option("--n", default=None, type=int,
              help="If passed, will check if a contour file has at least that many contours")
@click.option("--root", default=".", help="source root to the Horos dataset")
def check_existing_contours(contour_types: typing.List[str], root: str, n: Optional[int]):
    """currently terrible and unhelpful vaidation process... honestly using pytest was pretty king"""
    click.echo(f"checking contour types {contour_types} in root: {root}...")
    combined_info = get_combined_info(root)

    click.echo("the following contours need to be rechecked: ")
    # TODO: improve logging...
    # really, pytest for this kinda stuff wasnt half bad
    for contour_type in contour_types:
        eligible = combined_info[
            (combined_info["contour_type"] == contour_type) & (combined_info["location_contour"].notna())]
        eligible.loc[:, "valid"] = eligible.apply(
            lambda row: contour_is_valid(load_horos_contour(row["location_contour"], row["location_images"]), n),
            axis=1)
        click.echo(eligible[eligible["valid"] != True][["ID", "contour_path", "contour_type"]])

    click.echo("finished checking")


main.add_command(make_image_info_csv)
main.add_command(make_contour_info_csv)
main.add_command(rename_contour_files)
main.add_command(validate)
main.add_command(check)
check.add_command(check_failed)
check.add_command(check_unobserved)
check.add_command(check_existing_contours)

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""Top-level package for horos_io."""

__author__ = """Quang Anh Le Hong"""
__email__ = 'qa12_8@yahoo.de'
__version__ = '0.1.0'

from .cmr import load_lax_sequence, sort_SAX_by_y, load_cine_sequence, load_sax_sequence, get_image_info, \
    get_contour_info, get_combined_info

from .core import load_horos_contour, get_n_frames_from_seq_path, get_n_slices_from_seq_path, \
    load_horos_contour_unstructured



"""
module for working with horos images and contours. the dataformat is as follows:
the names for anonymous study and the sequences are inconsistent

ROOT |--- ID0
                 |---------   Anonymous_Study
                                    |------------ <LAX>
                                                    |----------- IM-0001-0001.dcm
                                                    |----------- IM-0001-0002.dcm
                                                    ...
                                    |------------ <SAX>
                                                    |----------- IM-0001-0001-0001.dcm
                                                    |----------- IM-0001-0001-0002.dcm
                                                    ...
                 |--- sax_lv_endo.xml
                 |--- sax_lv_epi.xml
                 ...
    |--- ID1
                 |--- ...
                 ...
    ...

not that it really matters;
for convenience, every location of contours and image seqs, will
    be stored in a .csv

CAVE: short axis stacks have a value for "basal_first", which needs to be passed
to the reading function to get the correct order of images

CAVE: currently does not respect the fact, that we could have non-CINEs in our data
"""
import os
import pathlib
from typing import Union

import numpy as np
import pandas as pd
import pydicom
from tqdm import tqdm

from horos_io import _config
from horos_io.core import get_n_frames_from_seq_path, \
    get_n_slices_from_seq_path, _get_name_from_template, get_contour_info_by_type
from ._utils import _to_str, get_seq_paths

Path = Union[os.PathLike, str]


def load_cine_sequence(path_to_sequence: Path) -> np.ndarray:
    """
    convenience function to let the algorithm decide, if it is lax or sax;

    a problem is, is that sequences can be repeated multiple times,
    the first code in IM-XXXX-0001.dcm or IM-XXXX-0001-0001.dcm denotes this;

    in our data we have only 1 sequence of however many repetitions there were
    Args:
        path_to_sequence:
    Returns:
    """
    p = path_to_sequence
    return load_lax_sequence(p) if get_n_slices_from_seq_path(p) == 1 else load_sax_sequence(p)


def load_sax_sequence(path_to_sequence: Path) -> np.ndarray:
    """

        for slices, they are sorted by the Y-component of the ImagePositionPatient Metadata of each slice;
        in theory, that value increases the more dorsal we go with respect to the patient.
        Ideally, the Apex should be the most ventral in the patient (inverted heart anatomy? scanning patient face down?)
    Args:
        path_to_sequence:
    Returns:
        a numpy array, where the contents of the array are pydicom.FileDataset of the images of shape
    (n_frames, n_slices)
    """
    n_frames = get_n_frames_from_seq_path(path_to_sequence)
    n_slices = get_n_slices_from_seq_path(path_to_sequence)

    result = np.array([[
        pydicom.dcmread(os.path.join(path_to_sequence,
                                     _get_name_from_template(path_to_sequence,
                                                             fr"IM-\d\d\d\d-00{_to_str(f + 1)}-00{_to_str(s + 1)}.dcm")))
        for s in range(n_slices)]
        for f in range(n_frames)])
    result = sort_SAX_by_y(result)
    return result


def load_lax_sequence(path_to_sequence: Path) -> np.ndarray:
    """
    returns a numpy array, where the contents of the array are pydicom.FileDataset of the images of shape
    (n_frames,)
    the indices correspond to the frame in the video
    Args:
        path_to_sequence:
    Returns:  np-array of the long axis images
    """
    n_frames = get_n_frames_from_seq_path(path_to_sequence)

    return np.array([
        pydicom.dcmread(os.path.join(path_to_sequence,
                                     _get_name_from_template(path_to_sequence, fr"IM-\d\d\d\d-00{_to_str(f + 1)}.dcm")))
        for f in range(n_frames)])


def _get_slice_type(path: Path) -> str:
    if "2ch" in str(path).lower():
        return "cine_2ch"
    elif "3ch" in str(path).lower():
        return "cine_3ch"
    elif "4ch" in str(path).lower():
        return "cine_4ch"
    elif "sax" in str(path).lower():
        return "cine_sa"
    else:
        raise ValueError(f"unknown slice type with path: {path}")


def sort_SAX_by_y(X: np.ndarray) -> np.ndarray:
    """
    Args:
        X: a 2D short axis stack;
        if dicom innolitics can be believed, we can assume that the Y-world axis increased the more dorsal
        you go with respect to the patient. therefore, in a SAX stack, the most apical slice should have the smallest y;
        we would be able to use this to sort the stack without relying on manually inputting the order;

        since we our SAX stack is of shape (n_frames, n_slices); need to transpose and retranspose
    """
    return np.array(
        sorted(X.T, key=lambda x: x[0].ImagePositionPatient[1])
    ).T


def get_image_info(root: Path) -> pd.DataFrame:
    """
    checks the root for any Horos Exported Sequences and returns a DataFrame with the information;
    should be in core; but since we currently also save the slice_type; which is cmr specific;
    it lies here for now...

    columns:
    ("seq_path",  <relative path starting from the root of the data>
     "location",  joins the rootdir with seq path
     "ID",        <Impression_Cmr{ID}>
     "slice_type", one of config.slice_types
     "n_frames",
     "n_slices",

    sequence can be loaded with load_sequence
    Args:
        root: root of the data; contents should be directories of Studies
    Returns:
        a DataFrame with the above information
    """
    result = pd.DataFrame()

    result["seq_path"] = get_seq_paths(root)

    result["ID"] = result["seq_path"].apply(lambda loc: pathlib.Path(loc).parts[0])  # ID is name of the first folder
    result["location"] = [os.path.normpath(os.path.join(root, seq_path))
                          for seq_path in get_seq_paths(root)]
    result["slice_type"] = result["location"].apply(_get_slice_type)
    result["n_frames"] = result["location"].apply(get_n_frames_from_seq_path)
    result["n_slices"] = result["location"].apply(get_n_slices_from_seq_path)
    result = result[["ID", "slice_type", "n_frames", "n_slices", "seq_path", "location"]]

    if result.shape[0] == 0:
        raise ValueError(f"could not find any images in {root}")

    return result.sort_values(by=["ID", "slice_type"])


def get_contour_info(root: Path) -> pd.DataFrame:
    """
    dataframe columns:
    ("ID": <Impression_Cmr{ID}>,
     "contour_path": relative path to the contour file, starting from root of the data,
     "contour_type": one of config.contour_types,
     "location": joins the root dir with the sequence path
     "slice_type": one of config.slice_types,
    )

    creates a csv file containing location of all contour files; or whether they exist at all.
    Args:
        root: root of the data; contents should be directories of Impression Studies
    Returns:
    """
    result = pd.concat([get_contour_info_by_type(root, contour_type)
                        for contour_type in tqdm(_config.contour_types)], ignore_index=True)
    result["slice_type"] = result["contour_type"].apply(lambda x: _get_slice_type(x))

    if result.shape[0] == 0:
        raise ValueError(f"could not find any contours in {root}")

    return result.sort_values(by=["ID", "slice_type", "contour_type"])


def get_combined_info(root: Path) -> pd.DataFrame:
    """
    Args:
        root: Horos Dataset root
    Returns:
        DataFrame, which contains the combined info you get from get_contour_info and get_image_info
        by merging with ID and slice_type
    """
    image_info = get_image_info(root)
    contour_info = get_contour_info(root)
    return pd.merge(contour_info[contour_info["location"].notna()],
                    image_info, on=["ID", "slice_type"], suffixes=("_contour", "_images"))


if __name__ == '__main__':
    pass



import pytest
from click.testing import CliRunner

from horos_io import cli

@pytest.fixture()
def runner():
    return CliRunner()

@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string
    del response


def test_command_line_interface(runner):
    """Test the CLI."""
    result = runner.invoke(cli.main)
    assert result.exit_code == 0

def test_make_image_info(runner, horos_test_root):
    result = runner.invoke(cli.main, ["image_info", "--root", horos_test_root])
    assert result.exit_code == 0





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['horos_io', 'tests']

package_data = \
{'': ['*'],
 'tests': ['horos_dummy/target_contour_info.csv',
           'horos_dummy/target_contour_info.csv',
           'horos_dummy/target_image_info.csv',
           'horos_dummy/target_image_info.csv']}

install_requires = \
['click',
 'matplotlib>=3.5.3,<4.0.0',
 'opencv-python>=4.6.0.66,<5.0.0.0',
 'pandas>=1.4.4,<2.0.0',
 'pydicom>=2.3.0,<3.0.0',
 'scipy>=1.9.1,<2.0.0',
 'tk>=0.1.0,<0.2.0',
 'tqdm>=4.64.1,<5.0.0']

entry_points = \
{'console_scripts': ['horos_io = horos_io.cli:main']}

setup_kwargs = {
    'name': 'horos-io',
    'version': '0.1.2',
    'description': 'small package to deal with data exported from Horos',
    'long_description': '========\nhoros_io\n========\n\n\n.. image:: https://img.shields.io/pypi/v/horos_io.svg\n        :target: https://pypi.python.org/pypi/horos_io\n\n.. image:: https://img.shields.io/travis/santomon/horos_io.svg\n        :target: https://travis-ci.com/santomon/horos_io\n\n.. image:: https://readthedocs.org/projects/horos-io/badge/?version=latest\n        :target: https://horos-io.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\n\nhandling .xml files when exported from horos; some dataset functionality\n\n\n* Free software: MIT\n* Documentation: https://horos-io.readthedocs.io.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n',
    'author': 'Quang Anh Le Hong',
    'author_email': 'qa12_8@yahoo.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/santomon/horos_io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
